unit FormDlgs;

{**********************************************************************}
{                                                                      }
{   FormDlgs                                                           }
{                                                                      }
{   base form used by all dialogs                                      }
{                                                                      }
{   Version     v0.96/12.05.05                                         }
{   Copyright   (c) 2003-2005 by DiK                                   }
{                                                                      }
{   Remarks:                                                           }
{   + all dialogs should be derived from this form                     }
{     - it centers the form on the main window (TSE),                  }
{     - validates string lists (ID -> TStrings),                       }
{     - handles OK, CANCEL and HELP messages,                          }
{     - handles some common error conditions,                          }
{     - defines some common utility routines and                       }
{     - defines an extended context menu for text input                }
{   + to fully utilize the context menu you must                       }
{     - assign it to the popup property of a combo box and             }
{     - assign FormContextPopup to the OnContextPopup event            }
{     - the application key is handled automatically                   }
{   + the extended popup menu currently works for combo boxes only     }
{   + and yes the popup handling is way to complicated :-(             }
{                                                                      }
{**********************************************************************}

interface

uses
  Windows, Messages, SysUtils, Classes, Forms, Controls, StdCtrls, Menus;

{----------------------------------------------------------------------}
{   common dialog base form                                            }
{----------------------------------------------------------------------}

type
  TFrmDlgs = class(TForm)
    BtnOK: TButton;
    BtnCancel: TButton;
    BtnHelp: TButton;
    MenuEdit: TPopupMenu;
    CmdUndo: TMenuItem;
    N1: TMenuItem;
    CmdCut: TMenuItem;
    CmdCopy: TMenuItem;
    CmdPaste: TMenuItem;
    CmdDelete: TMenuItem;
    N2: TMenuItem;
    CmdSelectAll: TMenuItem;
    N3: TMenuItem;
    MenuCharacter: TMenuItem;
    CmdBackspace: TMenuItem;
    CmdTabulator: TMenuItem;
    CmdNewline: TMenuItem;
    CmdReturn: TMenuItem;
    CmdFormFeed: TMenuItem;
    CmdEscape: TMenuItem;
    N4: TMenuItem;
    CmdCode: TMenuItem;
    procedure FormKeyDown(
      Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormContextPopup(
      Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure MenuEditPopup(Sender: TObject);
    procedure CmdClick(Sender: TObject);
    procedure CmdCharClick(Sender: TObject);
    procedure CmdCodeClick(Sender: TObject);
    procedure BtnHelpClick(Sender: TObject);
  private
    fLastError: Integer;
    procedure CenterForm;
    function GetActiveComboBox(var CB: TComboBox): Boolean;
  protected
    function ValidateStrings(ID: Integer): TStrings;
    procedure Error(Code: Integer);
    procedure InitComboBox(ComboBox: TComboBox; StrList: TStrings);
  public
    function Execute(X: Integer = 0; Y: Integer = 0): Integer;
    property LastError: Integer read fLastError;
  end;

{----------------------------------------------------------------------}
{   common resource strings                                            }
{----------------------------------------------------------------------}

resourcestring
  rsWarn            = 'Warning';
  rsError           = 'Error';
  rsErrorFormat     = 'Function %s: %s';
  rsInvalidID       = 'invalid list ID';
  rsInternalError   = 'internal error';

resourcestring
  rsDelAllHist      = 'Delete ALL history entries?';

{**********************************************************************}
{   implementation                                                     }
{**********************************************************************}

implementation

uses
  ClipBrd, Common, StrList, FormMsgBox, FormChar;

{$R *.dfm}

{----------------------------------------------------------------------}
{   find the edit control of a combo box (for sub-classing)            }
{----------------------------------------------------------------------}

var
  EditCtrl: THandle;

function EnumChildProc(Handle: HWND; LParam: Integer): BOOL; stdcall;
begin
  EditCtrl := Handle;
  Result := False;
end;

function GetEditCtrl(CB: TComboBox): THandle;
begin
  EditCtrl := 0;
  EnumChildWindows(CB.Handle, @EnumChildProc, 0);
  Result := EditCtrl;
end;

{----------------------------------------------------------------------}
{   internal helpers                                                   }
{----------------------------------------------------------------------}

procedure TFrmDlgs.CenterForm;
var
  Rect: TRect;
begin
  if GetWindowRect(Application.Handle, Rect) then
    with Rect do
    begin
      Self.Left := Left + (Right - Left - Self.Width) div 2;
      Self.Top := Top + (Bottom - Top - Self.Height) div 2;
    end;
end;

function TFrmDlgs.GetActiveComboBox(var CB: TComboBox): Boolean;
begin
  Result := ActiveControl is TComboBox;
  if Result then
    CB := TComboBox(ActiveControl)
  else
    CB := nil;
end;

{----------------------------------------------------------------------}
{   utility routines                                                   }
{----------------------------------------------------------------------}

function TFrmDlgs.ValidateStrings(ID: Integer): TStrings;
begin
  if not GetStringList(ID, Result) then Error(slInvalidID);
end;

procedure TFrmDlgs.Error(Code: Integer);
var
  S: string;
begin
  fLastError := Code;
  case Code of
    slInvalidID:    S := rsInvalidID;
    else            S := rsInternalError;
  end;
  S := Format(rsErrorFormat, [Caption, S]);
  if Self is TFrmMsgBox then
    Application.MessageBox(PChar(S), PChar(rsError), 0)
  else
    DlgMsgBox(rsError, S);
  Abort;
end;

procedure TFrmDlgs.InitComboBox(ComboBox: TComboBox; StrList: TStrings);
begin
  ComboBox.Items.AddStrings(StrList);
  if StrList.Count > 0 then ComboBox.Text := StrList[0];
end;

{----------------------------------------------------------------------}
{   display a dialog                                                   }
{----------------------------------------------------------------------}

function TFrmDlgs.Execute(X, Y: Integer): Integer;
begin
  if fLastError = slOK then
  begin
    if (X = 0) and (Y = 0) then
    begin
      CenterForm;
    end else
    begin
      Left := X;
      Top := Y;
    end;
    Result := ShowModal;
  end else
    Result := fLastError;
end;

{----------------------------------------------------------------------}
{   key handler                                                        }
{----------------------------------------------------------------------}

procedure TFrmDlgs.FormKeyDown(
  Sender: TObject; var Key: Word; Shift: TShiftState);

  procedure CheckCombo;
  var
    R: TRect;
    CB: TComboBox;
  begin
    if GetActiveComboBox(CB) and Assigned(CB.PopupMenu) then
    begin
      GetWindowRect(CB.Handle, R);
      CB.PopupMenu.Popup(R.Left + CB.Width div 2, R.Top + CB.Height div 2);
      Key := 0;
    end;
  end;

begin
  case Key of
    VK_F1:      BtnHelp.Click;
    VK_F10:     if Shift = [ssShift] then CheckCombo;
    VK_APPS:    CheckCombo;
  end;
end;

{----------------------------------------------------------------------}
{   extended context menu                                              }
{----------------------------------------------------------------------}

procedure TFrmDlgs.FormContextPopup(
  Sender: TObject; MousePos: TPoint; var Handled: Boolean);
begin
  if Sender is TComboBox then
    TComboBox(Sender).SetFocus;
end;

procedure TFrmDlgs.MenuEditPopup(Sender: TObject);
var
  CB: TComboBox;
  BE: Boolean;
begin
  if GetActiveComboBox(CB) then
  begin
    BE := CB.SelLength > 0;
    CmdUndo.Enabled := Boolean(SendMessage(GetEditCtrl(CB), EM_CANUNDO, 0, 0));
    CmdCut.Enabled := BE;
    CmdCopy.Enabled := BE;
    CmdDelete.Enabled := BE;
    CmdPaste.Enabled := Clipboard.HasFormat(CF_TEXT);
    CmdSelectAll.Enabled := CB.SelLength <> Length(CB.Text);
  end;
end;

procedure TFrmDlgs.CmdClick(Sender: TObject);
var
  CB: TComboBox;
begin
  if GetActiveComboBox(CB) then
    case (Sender as TMenuItem).Tag of
      1: SendMessage(GetEditCtrl(CB), EM_UNDO, 0, 0);
      2: CB.Perform(WM_CUT, 0, 0);
      3: CB.Perform(WM_COPY, 0, 0);
      4: CB.Perform(WM_PASTE, 0, 0);
      5: CB.Perform(WM_CLEAR, 0, 0);
      6: CB.SelectAll;
    end
end;

procedure TFrmDlgs.CmdCharClick(Sender: TObject);
var
  CB: TComboBox;
begin
  if GetActiveComboBox(CB) then
    CB.SelText := Char((Sender as TMenuItem).Tag);
end;

procedure TFrmDlgs.CmdCodeClick(Sender: TObject);
var
  CB: TComboBox;
  C: Integer;
  S: string;
begin
  if GetActiveComboBox(CB) then
  begin
    S := CB.SelText;
    if Length(S) > 0 then C := Ord(S[1]) else C := 42;
    C := GetSpecialChar(C);
    if C >= 0 then CB.SelText := Chr(C);
  end;
end;

{----------------------------------------------------------------------}
{   help response                                                      }
{----------------------------------------------------------------------}

procedure TFrmDlgs.BtnHelpClick(Sender: TObject);
begin
  ModalResult := idHelp;
end;

end.
