unit IncrSrch;

{======================================================================}
{                                                                      }
{   unit IncrSrch                                     Version 1.3      }
{                                                                      }
{   Implements incremental searching for assorted controls             }
{                                                                      }
{   Copyright       (c) 1997-2001 by DiK                               }
{                                                                      }
{   History                                                            }
{   v1.31/28.07.01  no changes                                         }
{   v1.30/11.12.98  added properties                                   }
{                   fixed handling of keyboard events                  }
{                   fixed deactivate when client looses focus          }
{                   fixed deactivate when form is deactivated          }
{   v1.20/22.07.98  added events                                       }
{                   added properties                                   }
{                   added deactivate when client looses focus          }
{                   added deactivate when form is deactivated          }
{                   clean up of source code                            }
{   v1.10/16.03.98  added ListViewIncrSearch                           }
{                   added StringGridIncrSearch                         }
{                   moved property editor to seperate unit             }
{                   moved HeadListIncrSearch to seperate unit          }
{                   fixed ClientClass (use TCustom... components)      }
{                   fixed gsColumns mode                               }
{                   clean up of source code                            }
{   v1.00/15.08.97  first version                                      }
{                                                                      }
{======================================================================}
{                                                                      }
{   Remark:                                                            }
{                                                                      }
{   Hooking form and application events is somewhat dangerous,         }
{   since it only works on a strictly last-in first-out basis.         }
{   Thus if you use multiple instances of the incremental search       }
{   classes on multiple forms, it is perhaps best to write your        }
{   own OnDeactivate handlers, both for the application and for        }
{   each form, and deactivate searching manually.                      }
{                                                                      }
{======================================================================}

{$I Switches.Inc}

{======================================================================}
{   INTERFACE                                                          }
{======================================================================}

interface

uses
  Windows, Messages, SysUtils, Classes,
  Menus, Forms, Controls, StdCtrls, ComCtrls, Grids;

{======================================================================}
{   type definitions                                                   }
{======================================================================}

const
  MaxGridColumn = 32;

type
  TGridColumns = Integer;
  TGridDirection = (gsRows,gsColumns);

{----------------------------------------------------------------------}
{   Exceptions                                                         }
{----------------------------------------------------------------------}

type
  EIncrSearchError = class(Exception);
  EInvalidSearchClient = class(EIncrSearchError);
  EInvalidGridColumn = class(EIncrSearchError);

{----------------------------------------------------------------------}
{   TCustomIncrSearch                                                  }
{----------------------------------------------------------------------}

type
  TCustomIncrSearch = class(TComponent)
  private
    // internal fields
    SearchAgain: Boolean;
    UndoStack: TList;
    ClientDeactivateApp: TNotifyEvent;
    ClientDeactivateForm: TNotifyEvent;
    ClientExit: TNotifyEvent;
    ClientKeyDown: TKeyEvent;
    ClientKeyPress: TKeyPressEvent;
    ClientMouseDown: TMouseEvent;
    // property fields
    FAgainKey: Word;
    FAgainShift: TShiftState;
    FActive: Boolean;
    FAutoDeactivate: Boolean;
    FAutoDisable: Boolean;
    FClient: TComponent;
    FEnabled: Boolean;
    FHookAppEvent: Boolean;
    FHookFormEvent: Boolean;
    FSearchText: String;
    FWantReturns: Boolean;
    FWantSpaces: Boolean;
    FOnActivate: TNotifyEvent;
    FOnDeactivate: TNotifyEvent;
    FOnSearchFailed: TNotifyEvent;
    FOnTextChanged: TNotifyEvent;
  private
    // setup and shutdown (client event hooks)
    procedure SwapAppEvents;
    procedure SwapFormEvents;
    procedure SwapClientEvents;
    function UpAndRunning: Boolean;
    // event functions
    procedure Activate;
    procedure Deactivate;
    procedure SearchFailed;
    procedure TextChanged;
    // property access functions (internal use only)
    procedure SetActive(Value: Boolean);
    procedure SetSearchText(const Value: String);
    // property access functions
    function GetAgainKey: TShortCut;
    procedure SetAgainKey(const Value: TShortCut);
    procedure SetClient(Value: TComponent);
    procedure SetEnabled(Value: Boolean);
    procedure SetHookAppEvent(const Value: Boolean);
    procedure SetHookFormEvent(const Value: Boolean);
    // incremental search functions
    procedure DoSearch(const NextChar: String);
    procedure UndoSearch;
    // client event hook functions
    procedure ISDeactivateApp(Sender: TObject);
    procedure ISDeactivateForm(Sender: TObject);
    procedure ISExit(Sender: TObject);
    procedure ISKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure ISKeyPress(Sender: TObject; var Key: Char);
    procedure ISMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  protected
    // setup and shutdown (run-time)
    procedure ClientChanged; virtual;
    procedure Loaded; override;
    procedure Notification(
      AComponent: TComponent; Operation: TOperation); override;
    // item access functions
    function ClientClass: TComponentClass; virtual; abstract;
    function FindText(
      var Index: Integer; Find: String): Boolean; virtual; abstract;
    function GetIndex: Integer; virtual; abstract;
    procedure SetIndex(Value: Integer); virtual; abstract;
  public
    // setup and shutdown
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    // run time interface
    procedure EndSearch;
    property Active: Boolean read FActive;
    property SearchText: String read FSearchText;
  published
    // design time interface
    property AutoDeactivate: Boolean read FAutoDeactivate write FAutoDeactivate;
    property AutoDisable: Boolean read FAutoDisable write FAutoDisable;
    property Client: TComponent read FClient write SetClient;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property HookAppEvent: Boolean read FHookAppEvent write SetHookAppEvent;
    property HookFormEvent: Boolean read FHookFormEvent write SetHookFormEvent;
    property SearchAgainKey: TShortCut read GetAgainKey write SetAgainKey;
    property WantReturns: Boolean read FWantReturns write FWantReturns;
    property WantSpaces: Boolean read FWantSpaces write FWantSpaces;
    property OnActivate: TNotifyEvent read FOnActivate write FOnActivate;
    property OnDeactivate: TNotifyEvent read FOnDeactivate write FOnDeactivate;
    property OnSearchFailed: TNotifyEvent
      read FOnSearchFailed write FOnSearchFailed;
    property OnTextChanged: TNotifyEvent
      read FOnTextChanged write FOnTextChanged;
  end;

{----------------------------------------------------------------------}
{   TListBoxIncrSearch                                                 }
{----------------------------------------------------------------------}

type
  TListBoxIncrSearch = class(TCustomIncrSearch)
  protected
    // item access functions
    function ClientClass: TComponentClass; override;
    function FindText(var Index: Integer; Find: String): Boolean; override;
    function GetIndex: Integer; override;
    procedure SetIndex(Value: Integer); override;
  end;

{----------------------------------------------------------------------}
{   TListViewIncrSearch                                                }
{----------------------------------------------------------------------}

type
  TListViewIncrSearch = class(TCustomIncrSearch)
  protected
    // item access functions
    function ClientClass: TComponentClass; override;
    function FindText(var Index: Integer; Find: String): Boolean; override;
    function GetIndex: Integer; override;
    procedure SetIndex(Value: Integer); override;
  end;

{----------------------------------------------------------------------}
{   TMemoIncrSearch                                                    }
{----------------------------------------------------------------------}

type
  TMemoIncrSearch = class(TCustomIncrSearch)
  protected
    // item access functions
    function ClientClass: TComponentClass; override;
    function FindText(var Index: Integer; Find: String): Boolean; override;
    function GetIndex: Integer; override;
    procedure SetIndex(Value: Integer); override;
  public
    // setup and shutdown
    constructor Create(AOwner: TComponent); override;
  end;

{----------------------------------------------------------------------}
{   TRichEditIncrSearch                                                }
{----------------------------------------------------------------------}

type
  TRichEditIncrSearch = class(TMemoIncrSearch)
  protected
    // item access functions
    function ClientClass: TComponentClass; override;
    function FindText(var Index: Integer; Find: String): Boolean; override;
  end;

{----------------------------------------------------------------------}
{   TCustomGridIncrSearch                                              }
{----------------------------------------------------------------------}

type
  TCustomGridIncrSearch = class(TCustomIncrSearch)
  private
    // property fields
    FColumns: TGridColumns;
    FDirection: TGridDirection;
  private
    // property access functions
    procedure SetColumns(Value: TGridColumns);
    procedure SetDirection(Value: TGridDirection);
  protected
    // setup and shutdown (run-time)
    procedure ClientChanged; override;
    // additional item access functions
    function GetCell(Col,Row: Integer): String; virtual; abstract;
    function GetCol: Integer; virtual; abstract;
    function GetRow: Integer; virtual; abstract;
    function GetColCount: Integer; virtual; abstract;
    function GetRowCount: Integer; virtual; abstract;
    procedure SetCol(Value: Integer); virtual; abstract;
    procedure SetRow(Value: Integer); virtual; abstract;
    // overloaded item access functions
    function FindText(var Index: Integer; Find: String): Boolean; override;
    function GetIndex: Integer; override;
    procedure SetIndex(Value: Integer); override;
  published
    // design time interface
    property Columns: TGridColumns read FColumns write SetColumns;
    property Direction: TGridDirection read FDirection write SetDirection;
  end;

{----------------------------------------------------------------------}
{   TStringGridIncrSearch                                              }
{----------------------------------------------------------------------}

type
  TStringGridIncrSearch = class(TCustomGridIncrSearch)
  protected
    // additional item access functions
    function ClientClass: TComponentClass; override;
    function GetCell(Col,Row: Integer): String; override;
    function GetCol: Integer; override;
    function GetRow: Integer; override;
    function GetColCount: Integer; override;
    function GetRowCount: Integer; override;
    procedure SetCol(Value: Integer); override;
    procedure SetRow(Value: Integer); override;
  end;

{======================================================================}
{   Registration                                                       }
{======================================================================}

procedure Register;

{======================================================================}
{   IMPLEMENTATION                                                     }
{======================================================================}

implementation

uses
  TypInfo;

{======================================================================}
{   resource strings                                                   }
{======================================================================}

resourcestring
  rsEnable  = 'Client must be assigned before enabling incremental search';
  rsClass   = 'Class of client must be %s or a descendant thereof';
  rsColumns = 'Client must be assigned before setting search columns';
  rsBounds  = 'Section column out of bounds';

{======================================================================}
{   internal types                                                     }
{======================================================================}

type
  TIntSet = set of 0..MaxGridColumn-1;

type // define this, so we can access the protected Items property
  TISCustomListView = class(TCustomListView)
  end;

{======================================================================}
{   utility function                                                   }
{   stolen from SysUtils and modified to work with DWORDs              }
{======================================================================}

procedure DivMod(
  Dividend: Integer; Divisor: Integer; var Result, Remainder: Integer);
asm
        PUSH    EBX
        MOV     EBX,EDX
        XOR     EDX,EDX
        DIV     EBX
        MOV     EBX,Remainder
        MOV     [ECX],EAX
        MOV     [EBX],EDX
        POP     EBX
end;

{======================================================================}
{   TCustomIncrSearch                                                  }
{======================================================================}

{----------------------------------------------------------------------}
{   setup and shutdown                                                 }
{----------------------------------------------------------------------}

constructor TCustomIncrSearch.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  UndoStack := TList.Create;
  ClientDeactivateApp := ISDeactivateApp;
  ClientDeactivateForm := ISDeactivateForm;
  ClientExit := ISExit;
  ClientKeyDown := ISKeyDown;
  ClientKeyPress := ISKeyPress;
  ClientMouseDown := ISMouseDown;
  SearchAgainKey := ShortCut(VK_F3,[]);
end;

destructor TCustomIncrSearch.Destroy;
begin
  if UpAndRunning and not (csDestroying in ComponentState) then EndSearch;
  UndoStack.Free;
  SwapAppEvents;
  SwapFormEvents;
  SwapClientEvents;
  inherited;
end;

procedure TCustomIncrSearch.ClientChanged;
begin
  if Assigned(FClient) and (csDesigning in ComponentState) then
  begin
    FEnabled := not FAutoDisable;
    FAutoDeactivate := not FAutoDisable;
  end;
end;

procedure TCustomIncrSearch.Loaded;
begin
  inherited;
  SwapClientEvents;
  SwapFormEvents;
  SwapAppEvents;
end;

procedure TCustomIncrSearch.Notification(
  AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (AComponent = FClient) and (Operation = opRemove) then
    if csDestroying in ComponentState then
      FClient := nil
    else
      SetClient(nil);
end;

function TCustomIncrSearch.UpAndRunning: Boolean;
begin
  Result := [csDesigning,csFixups,csLoading] * ComponentState = [];
end;

{----------------------------------------------------------------------}
{   setting client event hooks                                         }
{----------------------------------------------------------------------}
{   Hooking the client and its form the way we do isn't strictly       }
{   necessary, after all we do know the class of the intended client   }
{   and we do know that our clients have mouse and keyboard events.    }
{   On the other hand, the chosen method has its advantages. It is     }
{   independent of the class of the client and it uses the same        }
{   interface to hook events as the form designers do.                 }
{----------------------------------------------------------------------}
{   Reference: Ray Lischner, Secrets of Delphi 2                       }
{----------------------------------------------------------------------}

procedure TCustomIncrSearch.SwapAppEvents;
var
  Event: TNotifyEvent;
begin
  if FHookAppEvent and UpAndRunning then
  begin
    Event := Application.OnDeactivate;
    Application.OnDeactivate := ClientDeactivateApp;
    ClientDeactivateApp := Event;
  end;
end;

procedure SwapEvent(Comp: TComponent; Event: String; var Method: TMethod);
var
  CompMethod: TMethod;
  CompEvent: PPropInfo;
begin
  CompEvent := GetPropInfo(Comp.ClassInfo,Event);
  if Assigned(CompEvent) and (CompEvent.PropType^.Kind = tkMethod) then
  begin
    CompMethod := GetMethodProp(Comp,CompEvent);
    SetMethodProp(Comp,CompEvent,Method);
    Method := CompMethod;
  end;
end;

procedure TCustomIncrSearch.SwapFormEvents;
begin
  if Assigned(Client) and FHookFormEvent and UpAndRunning then
  begin
    SwapEvent(Client.Owner,'OnDeactivate',TMethod(ClientDeactivateForm));
  end;
end;

procedure TCustomIncrSearch.SwapClientEvents;
begin
  if Assigned(Client) and UpAndRunning then
  begin
    SwapEvent(Client,'OnExit',TMethod(ClientExit));
    SwapEvent(Client,'OnKeyDown',TMethod(ClientKeyDown));
    SwapEvent(Client,'OnKeyPress',TMethod(ClientKeyPress));
    SwapEvent(Client,'OnMouseDown',TMethod(ClientMouseDown));
  end;
end;

{----------------------------------------------------------------------}
{   event functions                                                    }
{----------------------------------------------------------------------}

procedure TCustomIncrSearch.Activate;
begin
  if Assigned(FOnActivate) then FOnActivate(Self);
end;

procedure TCustomIncrSearch.Deactivate;
begin
  if Assigned(FOnDeactivate) then FOnDeactivate(Self);
end;

procedure TCustomIncrSearch.SearchFailed;
begin
  if Assigned(FOnSearchFailed) then FOnSearchFailed(Self) else Beep;
end;

procedure TCustomIncrSearch.TextChanged;
begin
  if Assigned(FOnTextChanged) then FOnTextChanged(Self);
end;

{----------------------------------------------------------------------}
{   property access functions (internal use only)                      }
{----------------------------------------------------------------------}

procedure TCustomIncrSearch.SetActive(Value: Boolean);
begin
  if Value <> FActive then
  begin
    FActive := Value;
    if FActive then
      Activate
    else
    begin
      // DoAutoDisable, must NOT call SetEnabled (infinte loop)
      if FAutoDisable then FEnabled := False;
      Deactivate;
    end;
  end;
end;

procedure TCustomIncrSearch.SetSearchText(const Value: String);
begin
  FSearchText := Value;
  TextChanged;
end;

{----------------------------------------------------------------------}
{   property access functions                                          }
{----------------------------------------------------------------------}

function TCustomIncrSearch.GetAgainKey: TShortCut;
begin
  Result := ShortCut(FAgainKey,FAgainShift);
end;

procedure TCustomIncrSearch.SetAgainKey(const Value: TShortCut);
begin
  ShortCutToKey(Value,FAgainKey,FAgainShift);
end;

procedure TCustomIncrSearch.SetEnabled(Value: Boolean);
begin
  if not Assigned(FClient) and not (csLoading in ComponentState) then
    raise EInvalidSearchClient.Create(rsEnable);
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    if not FEnabled and UpAndRunning then EndSearch;
  end;
end;

procedure TCustomIncrSearch.SetClient(Value: TComponent);
begin
  if Assigned(Value) and not (Value is ClientClass) then
    raise EInvalidSearchClient.CreateFmt(rsClass,[ClientClass.ClassName]);
  if FClient <> Value then
  begin
    if UpAndRunning then EndSearch;     // stop searching
    SwapClientEvents;                   // unhook old client
    FClient := Value;                   // set new client
    SwapClientEvents;                   // hook new client
    ClientChanged;                      // notify self about change
    if not Assigned(FClient) then FEnabled := False;
  end;
end;

procedure TCustomIncrSearch.SetHookAppEvent(const Value: Boolean);
begin
  if FHookAppEvent <> Value then
  begin
    if UpAndRunning then EndSearch;
    SwapAppEvents;
    FHookAppEvent := Value;
    SwapAppEvents;
  end;
end;

procedure TCustomIncrSearch.SetHookFormEvent(const Value: Boolean);
begin
  if FHookFormEvent <> Value then
  begin
    if UpAndRunning then EndSearch;
    SwapFormEvents;
    FHookFormEvent := Value;
    SwapFormEvents;
  end;
end;

{----------------------------------------------------------------------}
{   search functions                                                   }
{----------------------------------------------------------------------}

procedure TCustomIncrSearch.DoSearch(const NextChar: String);
var
  Text: String;
  Index: Integer;
begin
  try

    // determine current position
    Index := GetIndex;
    if Index < 0 then
    begin
      Index := 0;
      SetIndex(0);
    end;

    // determine text to search and start position
    if NextChar = '' then
    begin
      Inc(Index);
      Text := SearchText;
    end else
      Text := SearchText + AnsiLowerCase(NextChar);

    // actually do search the text
    if FindText(Index,Text) then
    begin
      if NextChar <> '' then UndoStack.Add(Pointer(GetIndex));
      SetSearchText(Text);
      SetIndex(Index);
    end else
      SearchFailed;

  // this should NOT happen
  except
    EndSearch;
    raise;
  end;
end;

procedure TCustomIncrSearch.UndoSearch;
begin
  with UndoStack do
    if Active and (Count > 0) then
      try
        SetSearchText(Copy(SearchText,1,Length(SearchText)-1));
        SetIndex(Integer(Items[Count-1]));
      finally
        Delete(Count-1);
        if AutoDeactivate and (Count = 0) then EndSearch;
      end;
end;

procedure TCustomIncrSearch.EndSearch;
begin
  if Active then
  begin
    try
      SetSearchText('');
      SetIndex(-1);
      SetActive(False);
    finally
      UndoStack.Clear;
    end;
  end;
end;

{----------------------------------------------------------------------}
{   client event hooks                                                 }
{----------------------------------------------------------------------}

procedure TCustomIncrSearch.ISDeactivateApp(Sender: TObject);
begin
  EndSearch;
  if Assigned(ClientDeactivateApp) then ClientDeactivateApp(Sender);
end;

procedure TCustomIncrSearch.ISDeactivateForm(Sender: TObject);
begin
  EndSearch;
  if Assigned(ClientDeactivateForm) then ClientDeactivateForm(Sender);
end;

procedure TCustomIncrSearch.ISExit(Sender: TObject);
begin
  EndSearch;
  if Assigned(ClientExit) then ClientExit(Sender);
end;

procedure TCustomIncrSearch.ISKeyDown(
  Sender: TObject; var Key: Word; Shift: TShiftState);
const
  VK_0 = Word('0'); // Borland never defines these for Pascal :-(
  VK_9 = Word('9');
  VK_A = Word('A');
  VK_Z = Word('Z');
const
  TypeableKeys = [VK_0..VK_9,VK_A..VK_Z,$BA..$C0,$DB..$E4,$E6,$E9..$F5];
begin
  if Enabled then
  begin

    // handle search again key
    SearchAgain := Active and (Key = FAgainKey) and (Shift = FAgainShift);
    if SearchAgain then
    begin
      DoSearch('');
      Key := 0;
      Exit;
    end;

    // trap typeable keys
    if (Key in TypeableKeys) or WantSpaces and (Key = VK_SPACE) then
    begin
      Exit;
    end;

    // handle special search keys
    if Active then
    begin
      case Key of
        VK_SHIFT,VK_CONTROL:
          Exit;
        VK_BACK:
          begin
            UndoSearch;
            Key := 0;
            Exit;
          end;
        VK_ESCAPE:
          begin
            EndSearch;
            Key := 0;
            Exit;
          end;
        VK_RETURN:
          begin
            EndSearch;
            if WantReturns then
            begin
              Key := 0;
              Exit;
            end;
          end;
        else
          EndSearch;
      end;
    end;
  end;

  // everything that falls through, must be handled by the client
  if Assigned(ClientKeyDown) then ClientKeyDown(Sender,Key,Shift);
end;

procedure TCustomIncrSearch.ISKeyPress(Sender: TObject; var Key: Char);
begin
  if Enabled then
  begin

    // handle typeable keys
    if (Key > ' ') or WantSpaces and (Key = ' ') then
    begin
      SetActive(True);
      DoSearch(Key);
      Key := #0;
      Exit;
    end;

    if Active then
    begin

      // trap search again key
      if SearchAgain then
      begin
        Key := #0;
        Exit;
      end;

      // trap special search keys
      case Key of
        #8,#27: // backspace, escape
          begin
            Key := #0;
            Exit;
          end;
        #13: // return
          if WantReturns then
          begin
            Key := #0;
            Exit;
          end;
      end;

    end;
  end;

  // everything that falls through, must be handled by the client
  if Assigned(ClientKeyPress) then ClientKeyPress(Sender,Key);
end;

procedure TCustomIncrSearch.ISMouseDown(
  Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  EndSearch;
  if Assigned(ClientMouseDown) then ClientMouseDown(Sender,Button,Shift,X,Y);
end;

{======================================================================}
{   TListBoxIncrSearch                                                 }
{======================================================================}

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TListBoxIncrSearch.ClientClass: TComponentClass;
begin
  ClientClass := TCustomListBox;
end;

function TListBoxIncrSearch.FindText(var Index: Integer; Find: String): Boolean;
begin
  with TCustomListBox(Client) do
  begin
    while Index < Items.Count do
    begin
      if Pos(Find,AnsiLowerCase(Items[Index])) > 0 then
      begin
        Result := True;
        Exit;
      end;
      Inc(Index);
    end;
    Result := False;
  end;
end;

function TListBoxIncrSearch.GetIndex: Integer;
begin
  Result := TCustomListBox(Client).ItemIndex;
end;

procedure TListBoxIncrSearch.SetIndex(Value: Integer);
begin
  if Value >= 0 then TCustomListBox(Client).ItemIndex := Value;
end;

{======================================================================}
{   TListViewIncrSearch                                                }
{======================================================================}

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TListViewIncrSearch.ClientClass: TComponentClass;
begin
  ClientClass := TCustomListView;
end;

function TListViewIncrSearch.FindText(var Index: Integer; Find: String): Boolean;
begin
  with TISCustomListView(Client) do
  begin
    while Index < Items.Count do
    begin
      if Pos(Find,AnsiLowerCase(Items[Index].Caption)) > 0 then
      begin
        Result := True;
        Exit;
      end;
      Inc(Index);
    end;
    Result := False;
  end;
end;

function TListViewIncrSearch.GetIndex: Integer;
begin
  with TISCustomListView(Client) do
    if Assigned(ItemFocused) then
      Result := ItemFocused.Index else Result := -1;
end;

procedure TListViewIncrSearch.SetIndex(Value: Integer);
begin
  if Value >= 0 then
    with TISCustomListView(Client).Items[Value] do
    begin
      Focused := True;
      Selected := True;
      MakeVisible(False);
    end;
end;

{======================================================================}
{   TMemoIncrSearch                                                    }
{======================================================================}

{----------------------------------------------------------------------}
{   setup and shutdown                                                 }
{----------------------------------------------------------------------}

constructor TMemoIncrSearch.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAutoDisable := True;
  FWantReturns := True;
  FWantSpaces := True;
end;

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TMemoIncrSearch.ClientClass: TComponentClass;
begin
  ClientClass := TCustomMemo;
end;

function TMemoIncrSearch.FindText(var Index: Integer; Find: String): Boolean;
var
  Item: String;
  Line,StartCol,FindPos: Integer;
begin
  with TCustomMemo(Client) do
  begin
    Line := Perform(EM_LINEFROMCHAR,Index,0);
    StartCol := Index - Perform(EM_LINEINDEX,Line,0);
    while Line < Lines.Count do
    begin
      Item := AnsiLowerCase(Lines[Line]);
      if StartCol > 0 then
        Item := Copy(Item,StartCol+1,Length(Item)-StartCol);
      FindPos := Pos(Find,Item);
      if FindPos > 0 then
      begin
        Index := Perform(EM_LINEINDEX,Line,0) + StartCol + FindPos - 1;
        Result := True;
        Exit;
      end;
      Inc(Line);
      StartCol := 0;
    end;
    Result := False;
  end;
end;

function TMemoIncrSearch.GetIndex: Integer;
begin
  Result := TCustomMemo(Client).SelStart;
end;

procedure TMemoIncrSearch.SetIndex(Value: Integer);
begin
  if Value >= 0 then
    with TCustomMemo(Client) do
    begin
      SelStart := Value;
      SelLength := Length(SearchText);
    end;
end;

{======================================================================}
{   TRichEditIncrSearch                                                }
{======================================================================}

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TRichEditIncrSearch.ClientClass: TComponentClass;
begin
  ClientClass := TCustomRichEdit;
end;

function TRichEditIncrSearch.FindText(var Index: Integer; Find: String): Boolean;
var
  SearchLength: Integer;
begin
  with TCustomRichEdit(Client) do
  begin
    SearchLength := Perform(WM_GETTEXTLENGTH,0,0) - Index;
    Index := FindText(Find,Index,SearchLength,[]);
    Result := Index >= 0;
  end;
end;

{======================================================================}
{   TCustomGridIncrSearch                                              }
{======================================================================}

{----------------------------------------------------------------------}
{   setup and shutdown                                                 }
{----------------------------------------------------------------------}

procedure TCustomGridIncrSearch.ClientChanged;
begin
  inherited;
  if not Assigned(Client) then FColumns := 0;
end;

{----------------------------------------------------------------------}
{   property access functions                                          }
{----------------------------------------------------------------------}

procedure TCustomGridIncrSearch.SetColumns(Value: TGridColumns);
begin
  if not (csLoading in ComponentState) then
  begin
    if not Assigned(Client) then
      raise EInvalidSearchClient.Create(rsColumns);
    if Value >= 1 shl GetColCount then
      raise EInvalidGridColumn.Create(rsBounds);
  end;
  FColumns := Value;
end;

procedure TCustomGridIncrSearch.SetDirection(Value: TGridDirection);
begin
  EndSearch;
  FDirection := Value;
end;

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TCustomGridIncrSearch.FindText(
  var Index: Integer; Find: String): Boolean;
var
  Row,Col,RowCount,ColCount: Integer;

  function Found: Boolean;
  begin
    Result :=
      ((FColumns = 0) or (Col in TIntSet(FColumns)))
        and (Pos(Find,AnsiLowerCase(GetCell(Col,Row))) > 0);
  end;

begin
  Result := True;
  RowCount := GetRowCount;
  ColCount := GetColCount;
  if FDirection = gsRows then
  begin
    DivMod(Index,ColCount,Row,Col);
    while Row < RowCount do
    begin
      while Col < ColCount do
      begin
        if Found then
        begin
          Index := Row * ColCount + Col;
          Exit;
        end;
        Inc(Col);
      end;
      Inc(Row);
      Col := 0;
    end
  end else // FDirection = gsColumns
  begin
    DivMod(Index,RowCount,Col,Row);
    while Col < ColCount do
    begin
      while Row < RowCount do
      begin
        if Found then
        begin
          Index := Col * RowCount + Row;
          Exit;
        end;
        Inc(Row);
      end;
      Inc(Col);
      Row := 0;
    end;
  end;
  Result := False;
end;

function TCustomGridIncrSearch.GetIndex: Integer;
begin
  if FDirection = gsRows then
    Result := GetColCount * GetRow + GetCol
  else
    Result := GetRowCount * GetCol + GetRow;
end;

procedure TCustomGridIncrSearch.SetIndex(Value: Integer);
var
  Row,Col: Integer;
begin
  if Value >= 0 then
  begin
    if FDirection = gsRows then
      DivMod(Value,GetColCount,Row,Col)
    else
      DivMod(Value,GetRowCount,Col,Row);
    SetRow(Row);
    SetCol(Col);
  end;
end;

{======================================================================}
{   TStringGridIncrSearch                                              }
{======================================================================}

{----------------------------------------------------------------------}
{   item access overrides                                              }
{----------------------------------------------------------------------}

function TStringGridIncrSearch.ClientClass: TComponentClass;
begin
  ClientClass := TStringGrid;
end;

function TStringGridIncrSearch.GetCell(Col,Row: Integer): String;
begin
  Inc(Col,TStringGrid(Client).FixedCols);
  Inc(Row,TStringGrid(Client).FixedRows);
  Result := TStringGrid(Client).Cells[Col,Row];
end;

function TStringGridIncrSearch.GetCol: Integer;
begin
  with TStringGrid(Client) do Result := Col - FixedCols;
end;

function TStringGridIncrSearch.GetRow: Integer;
begin
  with TStringGrid(Client) do Result := Row - FixedRows;
end;

function TStringGridIncrSearch.GetColCount: Integer;
begin
  with TStringGrid(Client) do Result := ColCount - FixedCols;
end;

function TStringGridIncrSearch.GetRowCount: Integer;
begin
  with TStringGrid(Client) do Result := RowCount - FixedRows;
end;

procedure TStringGridIncrSearch.SetCol(Value: Integer);
begin
  with TStringGrid(Client) do Col := Value + FixedCols;
end;

procedure TStringGridIncrSearch.SetRow(Value: Integer);
begin
  with TStringGrid(Client) do Row := Value + FixedRows;
end;

{======================================================================}
{   Registration                                                       }
{======================================================================}

procedure Register;
begin
  RegisterComponents('DiK',
    [TListBoxIncrSearch,
     TListViewIncrSearch,
     TMemoIncrSearch,
     TRichEditIncrSearch,
     TStringGridIncrSearch]);
end;

end.
