/****************************************************************************\

    Grep.S

    Grep command to search text and/or files.

    Version         v1.22/01.04.03
    Copyright       (c) 1997-2003 by DiK

    Overview:

    This macro searches for a string or a regular expression.
    These may be located within the files loaded into memory
    or within files located on disk.

    Keys:       (none)

    Command Line Format:

    Grep  [disk|refresh]

    where:
        disk    force disk searching mode
        refresh re-run grep with previous options

    History
    v1.22/01.04.03  fixed dialog exit code
    v1.21/09.04.02  fixed version checking
    v1.20/25.04.01  adaption to v3.0 of editor
    v1.14/05.08.99  bug fix, infinite loop (finally!)
    v1.13/27.04.99  bug fix, names only when scanninf mem files
    v1.12/14.08.98  bug fix, changed include paths
    v1.11/16.01.98  bug fix
    v1.10/27.11.97  enhanced scope changes at startup
    v1.00/07.04.97  first version

\****************************************************************************/

/****************************************************************************\
    global constants
\****************************************************************************/

#include "dialog.si"
#include "msgbox.si"

/****************************************************************************\
    global variables
\****************************************************************************/

integer ref_count                       // count of recursive calls

integer dirstack                        // working buffer (recursion stack)
integer filelist                        // ditto (dir listings)
integer viewbuff                        // ditto (results)
integer tempbuff                        // ditto (results of text search)

integer verbose                         // parameter (verbosity)
integer subdirs                         // ditto (recurse subdirs)
integer mem_only                        // ditto (search files in memory)
integer curr_only                       // ditto (search only current file)
integer show_text                       // ditto (list filenames only)

string opts[8]                          // find options
string expr[80]                         // text to find
string wild[255]                        // files to find
string excl[255]                        // files to exclude
string root[255]                        // list of root folders

integer total_files                     // counter (files found)
integer total_count                     // ditto (occurences found)

integer curr_line                       // position within file at startup
integer curr_buff                       // id of file at startup
string curr_name[255]                   // name of file at startup

#ifdef WIN32
integer saved_mode                      // screen info
#endif
integer status_attr                     // screen info
integer screen_width                    // ditto
integer screen_height                   // ditto

string msg_esc[] = "Press <Escape> to stop searching..."
string msg_abort[] = "File: *** STOP *** Aborted by User *** STOP ***" + Chr(0)

/****************************************************************************\
    shared code
\****************************************************************************/

#include "scrun.si"

/****************************************************************************\
    dos compatibility
\****************************************************************************/

#ifndef WIN32

integer proc MsgBox(string cap, string msg, integer btn)
    Message(msg)
    return(0)
end

string proc QuotePath( string path )
    return(path)
end

string proc AddTrailingSlash( string path )
    return(iif( path[Length(path)] == "\", path, path + '\' ))
end

string proc PBName()
    string name[16]

    name = RTrim(GetText(11,8))
    if CurrChar(21) <> Asc(" ")
        name = name + "." + RTrim(GetText(21,3))
    endif
    return(name)
end

string DTA[64]

integer proc FindFirstFile( string filename, integer attributes )
    integer rc

    SetDTA(DTA)
    rc = FindFirst(filename,attributes)
    if rc
        return(1)
    endif
    return(-1)
end

integer proc FindNextFile()
    SetDTA(DTA)
    return(FindNext())
end

integer proc FFAttribute()
    return(Asc(DTA[22]))
end

string proc FFName()
    integer p
    string name[14]

    name = DTA[31:13]
    p = Pos(Chr(0),name)
    return(name[1..p-1])
end

#endif

/****************************************************************************\
    helper routines
\****************************************************************************/

integer flag_aborted

integer proc Aborted()
    if not flag_aborted
        flag_aborted = KeyPressed() and GetKey() == <Escape>
        if flag_aborted
            AddLine(msg_abort,viewbuff)
        endif
    endif
    return(flag_aborted)
end

proc TranslateWildCards( var string wc )
    integer bid

    bid = GotoBufferId(tempbuff)
    AddLine(wc)
    lReplace("{[.^$+@#{}]}","\\\1","cgxn")      // escape special characters
    lReplace(";","$}|{^","cgn")                 // group multiple filenames
    lReplace("*",".*","cgn")                    // replace dos wildcards
    lReplace("?",".","cgn")                     // with regular expression
    wc = "{^" + GetText(1,CurrLineLen()) +"$}"
    EmptyBuffer()
    GotoBufferId(bid)
end

string proc TruncName( string name )
    integer len = screen_width - 12

    if Length(name) > len
        len = Length(name) - len
        return(
            Format(name[1..3],"...",name[len..Length(name)])
        )
    endif
    return(name)
end

/****************************************************************************\
    screen routine
\****************************************************************************/

proc InitScreen()
    if verbose
#ifdef WIN32
        saved_mode = Set(CurrVideoMode,_25_LINES_)
#endif
        status_attr = Query(StatusLineAttr)
        screen_width = Query(ScreenCols)
        screen_height = Query(ScreenRows)
        if PopWinOpen(1,1,screen_width,screen_height,1,"",0)
            Set(Attr,0)
            ClrScr()
        else
            verbose = FALSE
        endif
        VGotoXY(0,screen_height-1)
        Set(Attr,status_attr)
        PutLine(msg_esc,screen_width)
        VHomeCursor()
    else
        status_attr = Query(StatusLineAttr)
        screen_width = Query(ScreenCols)
        screen_height = Query(ScreenRows)
        VGotoXY(1,screen_height)
        Set(Attr,status_attr)
        PutLine(msg_esc,screen_width)
    endif
end

proc DoneScreen()
    if verbose
        PopWinClose()
    endif
#ifdef WIN32
    if saved_mode
        Set(CurrVideoMode,saved_mode)
    endif
#endif
    UpdateDisplay(_HELPLINE_REFRESH_|_STATUS_LINE_REFRESH_)
end

/****************************************************************************\
    add results to view buffer and screen
\****************************************************************************/

proc ShowFileName( string text )
    if verbose
        Set(Attr,Color(Bright Yellow on Black))
        WriteLine(text)
        Set(Attr,Color(Bright White on Black))
    else
        Message(text)
    endif
end

proc SaveTextLine( integer line, string text )
    string buff[255]

    buff = Format( line:6, ": ", text )
    AddLine(buff,tempbuff)
    if verbose
        WriteLine(buff[1..screen_width-2])
    endif
end

proc OutputFileName( integer hits, string name )
    integer len = Length(name)
    string line[255]

    len = Max(43-len,1)

#if EDITOR_VERSION >= 0x3000
    line = Format( 'File: "', name, '" ',
        iif(hits, Format(" ":len, hits, " occurrences found" ), ""))
#else
    line = Format( "File: ", name, Chr(0),
        iif(hits, Format(" ":len, hits, " occurrences found" ), ""))
#endif

    AddLine(line,viewbuff)
end

proc OutputTextBlock()
    integer bid, ilba

    bid = GotoBufferId(tempbuff)
    MarkLine(1,NumLines())
    GotoBufferId(viewbuff)
    ilba = Set(InsertLineBlocksAbove,FALSE)
    MoveBlock()
    Set(InsertLineBlocksAbove,ilba)
    UnmarkBlock()
    EndFile()
    BegLine()
    GotoBufferId(bid)
end

/****************************************************************************\
    search text inside of files
\****************************************************************************/

proc SearchText()
    integer count = 0

    if not lFind(expr,opts)
        return()
    endif
    repeat
        EndLine()
        count = count + 1
        if show_text
            SaveTextLine(CurrLine(),GetText(1,CurrLineLen()))
        endif
    until not lRepeatFind()
    if not Aborted()
        OutputFileName(count,CurrFilename())
        if show_text
            OutputTextBlock()
        endif
        total_files = total_files + 1
        total_count = total_count + count
    endif
end

/****************************************************************************\
    search matching files
\****************************************************************************/

proc SearchFiles( string path )
    integer bid, pff
    string name[255]

    // build list of files

    GotoBufferId(filelist)
    EmptyBuffer()
    pff = Set(PickFileFlags,_NONE_)
#ifdef WIN32
    BuildPickBufferEx(path+"*",_NORMAL_)
#else
    BuildPickBuffer(path+"*.*",_NORMAL_)
#endif
    Set(PickFileFlags,pff)

    // translate pick buffer to list of filenames

    if NumLines() == 0
        return()
    endif
    BegFile()
    repeat
        name = Upper(PBName())
        DelToEol()
        InsertText(name)
        BegLine()
    until not Down()

    // filter list of filenames for included names

    BegFile()
    loop
        if lFind(wild,"cgx")
            if not Down()
                break
            endif
        else
            KillLine()
            if CurrLine() > NumLines()
                break
            endif
        endif
    endloop

    // filter list of filenames for excluded names

    if NumLines() == 0
        return()
    endif
    if Length(excl)
        while lFind(excl,"gx")
            KillLine()
        endwhile
    endif

    // search remaining files for text

    if NumLines() == 0
        return()
    endif
    BegFile()
    repeat
        name = path + GetText(1,CurrLineLen())
        ShowFileName(TruncName(name))
        if Length(expr)
            bid = GetBufferId(name)
            if bid                                  // file already loaded
                GotoBufferId(bid)
                SearchText()
            else                                    // load and quit file
                bid = EditFile(QuotePath(name),_DONT_PROMPT_)
                if bid
                    SearchText()
                    AbandonFile(bid)
                endif
            endif
            GotoBufferId(filelist)
        else
            total_files = total_files + 1
            OutputFileName(0,name)
        endif
    until Aborted() or not Down()
end

/****************************************************************************\
    push folders onto recursion stack
\****************************************************************************/

proc SearchFolders( string path )
    integer handle
    string name[255]

#ifdef WIN32
    handle = FindFirstFile(path+"*",_DIRECTORY_)
#else
    handle = FindFirstFile(path+"*.*",_DIRECTORY_)
#endif
    if handle == -1
        return()
    endif
    repeat
        if FFAttribute() & _DIRECTORY_
            name = FFName()
            if not (name in ".","..")
                InsertLine(path+name)
            endif
        endif
#ifdef WIN32
    until not FindNextFile(handle,_DIRECTORY_)
    FindFileClose(handle)
#else
    until not FindNextFile()
#endif
end

/****************************************************************************\
    grep disk
\****************************************************************************/

proc GrepDisk()
    integer k
    string path[255]

    InitScreen()
    TranslateWildCards(wild)
    TranslateWildCards(excl)

    for k = 1 to NumTokens(root,";")
        path = GetToken(root,";",k)
        path = SplitPath(ExpandPath(path),_DRIVE_|_PATH_)
        AddLine(path)
    endfor

    while not Aborted() and NumLines()
        BegFile()
        path = AddTrailingSlash(GetText(1,CurrLineLen()))
        KillLine()
        Message(TruncName(path))
        SearchFiles(path)
        GotoBufferId(dirstack)
        BegFile()
        if subdirs
            Message(TruncName(path))
            SearchFolders(path)
        endif
    endwhile

    DoneScreen()
end

/****************************************************************************\
    grep memory
\****************************************************************************/

proc GrepMem()
    integer num_files

    GotoBufferId(curr_buff)
    InitScreen()
    if curr_only
        num_files = 1
    else
        num_files = NumFiles() + (BufferType() <> _NORMAL_)
    endif
    do num_files times
        ShowFileName(TruncName(CurrFilename()))
        SearchText()
        NextFile()
    enddo
    DoneScreen()
    GotoBufferId(curr_buff)
end

/****************************************************************************\
    grep engine
\****************************************************************************/

integer proc Grep()
    integer line
    integer rc = FALSE

    // save editor state

    PushBlock()
    PushPosition()
    curr_line = CurrLine()
    curr_buff = GetBufferId()
    curr_name = CurrFilename()

    // do search

    tempbuff = CreateTempBuffer()
    viewbuff = CreateTempBuffer()
    filelist = CreateTempBuffer()
    dirstack = CreateTempBuffer()

    if dirstack and filelist and viewbuff and tempbuff
        UpdateDisplay()
        if mem_only or curr_only
            GrepMem()
        else
            GrepDisk()
        endif
        rc = total_files or total_count
    else
        Warn("Cannot allocate working buffers")
    endif

    AbandonFile(dirstack)
    AbandonFile(filelist)
    AbandonFile(tempbuff)

    // position view buffer

    GotoBufferId(viewbuff)
    if rc and lFind("File: "+curr_name,"g^")
        repeat
            if not Down()
                break
            endif
            line = Val(GetText(1,6))
        until line == 0 or line >= curr_line
    endif

    // restore editor state

    GotoBufferId(curr_buff)
    PopPosition()
    PopBlock()
    UpdateDisplay(_ALL_WINDOWS_REFRESH_)

    return(rc)
end

/****************************************************************************\
    parameter handling
\****************************************************************************/

proc CheckNonFindParam( string param, var integer switch )
    switch = Pos(param,opts)
    if switch
        opts = DelStr(opts,switch,1)
        switch = TRUE
    endif
end

proc GetDialogParameters()
    integer dummy

    // get dialog parameters

    expr = GetHistoryStr(GetFreeHistory("DlgGrep:Expression"),1)
    wild = GetHistoryStr(GetFreeHistory("DlgGrep:Filename"),1)
    excl = GetHistoryStr(GetFreeHistory("DlgGrep:Exclude"),1)
    root = GetHistoryStr(GetFreeHistory("DlgGrep:Pathname"),1)
    opts = GetHistoryStr(GetFreeHistory("DlgGrep:Options"),1)

    // translate parameters for grep

    CheckNonFindParam("n",show_text)
    CheckNonFindParam("s",subdirs)
    CheckNonFindParam("v",verbose)
    CheckNonFindParam("d",dummy)
    CheckNonFindParam("m",mem_only)
    CheckNonFindParam("c",curr_only)

    if mem_only or curr_only
        show_text = FALSE
        subdirs = FALSE
    endif

    show_text = not show_text
    opts = "g" + opts
    wild = Upper(wild)
    excl = Upper(excl)

    if wild == ""
#ifdef WIN32
        wild = "*"
#else
        wild = "*.*"
#endif
    endif
    if root == ""
        root = "."
    endif
end

/****************************************************************************\
    screen refresh
\****************************************************************************/

proc ClearScreen()
    if NumFiles() == 0
        Set(Attr,Query(TextAttr))
        ClrScr()
    endif
end

/****************************************************************************\
    main program
\****************************************************************************/

proc Main()
    integer rc
    integer not_refresh

set(break,on)
    // startup

#ifdef WIN32
    saved_mode = 0
#endif

    ref_count = ref_count + 1
    not_refresh = Query(MacroCmdLine) <> "refresh"

    // run dialog and get parameters

    if not_refresh
        rc = ExecDialog("dlggrep "+Query(MacroCmdLine))
        if not rc or Val(Query(MacroCmdLine)) <> ID_OK
            goto common_exit
        endif
    endif
    GetDialogParameters()

    // run grep and display results

    ClearScreen()
    if Grep()
        Message("found ",total_files," file(s)",
            iif(total_count,
                Format(" and ",total_count," occurence(s) of expression"),
                ""))
        if Query(ViewFindsId)
            AbandonFile(Query(ViewFindsId))
        endif
        Set(ViewFindsId,viewbuff)
        if not_refresh
            ExecDialog("viewgrep")
        endif
    else
        ClearScreen()
        MsgBox("Grep", "No matching files and/or text found", 0)
        AbandonFile(viewbuff)
        UpdateDisplay(_ALL_WINDOWS_REFRESH_)
    endif

    // cleanup

common_exit:

    ref_count = ref_count - 1
    if ref_count == 0
        PurgeMacro(CurrMacroFilename())
    endif

end
