/****************************************************************************\

    ViewGrep.S

    Display module for Grep macro.

    Version         v1.30/17.04.02
    Copyright       (c) 1997-2002 by DiK

    Overview:

    This macro manages the view buffer of the grep command.

    Keys:       (none)

    History

    v1.30/17.04.02  adaption to version 4.0 of editor
                    fixed mouse wheel support
                    fixed version checking
    v1.21/15.01.02  added mouse wheel support (needs mswheel)
    v1.20/25.04.01  adaption to v3.0 of editor
    v1.12/06.12.00  centered help
    v1.11/16.11.99  fixed idle update counting
    v1.10/05.08.99  fixed and enhanced context window
    v1.01/08.12.97  bug fix
    v1.00/07.04.97  first version

\****************************************************************************/

/****************************************************************************\
    global constants
\****************************************************************************/

#include "dialog.si"

/****************************************************************************\
    termination codes
\****************************************************************************/

constant STOP_QUIT = 1
constant STOP_GOTO = 2
constant STOP_EDIT = 3
constant STOP_GREP = 4
constant STOP_CTXT = 5

/****************************************************************************\
    configuration values
\****************************************************************************/

constant IDLE_UPDATE_DELAY = 9
constant DOUBLE_CLICK_DELAY = 9

/****************************************************************************\
    global variables
\****************************************************************************/

integer mouse_ticks                     // time passed since last mouse click
integer idle_ticks                      // context window (time for painting)
integer idle_paint                      // ditto (delayed painting)
integer ctxt_start                      // ditto (first line)
integer ctxt_lines                      // ditto (window size)
integer ctxt_offset                     // ditto (line offset)
integer hist_lines                      // ditto (history number)

string ResultLine[] = "^ *[0-9]#: "

#if EDITOR_VERSION >= 0x3000
string HeaderLine[] = '^File: "?{.+}"? #[0-9]# occurrences found{ on [0-9]# lines}?$'
//== pre version 3.0 ==
//string HeaderLine[] = '^File: "?{.+}"? #[0-9]# occurrences found$'
//== ==
#else
string HeaderLine[] = '^File: {[~ ]#}'
//?? pre version 2.50e ??
//string HeaderLine[] = '^File: {[~' + Chr(0) + ']#}'
//?? ??
#endif

/****************************************************************************\
    dos compatibility
\****************************************************************************/

#ifndef WIN32

string proc QuotePath( string path )
    return(path)
end

#endif

/****************************************************************************\
    shared code
\****************************************************************************/

#include "schelp.si"

/****************************************************************************\
    help screen
\****************************************************************************/

constant HLP_WIDTH = 66

helpdef ViewFindsHelp
    title = "Help on ViewFinds"
    width = HLP_WIDTH
    height = 21
    y = 3

    ''
    ' Summary of available commands'
    ''
    ' Enter                go to current line within current file'
    ''
    ' Alt L                load current file into memory'
    ' Alt E                edit the view finds list'
    ' Alt H                remove everything except header lines'
    ' Alt R                delete the current entry'
    ' Alt T                delete all entries except the current one'
    ' Alt C                change height of context window'
    ''
    ' F5                   refresh the list'
    ''
    ' Ctrl CursorUp        goto previous header line'
    ' Ctrl CursorDown      goto next header line'
    ''
    ' Grey-                scroll context window up one line'
    ' Grey+                scroll context window down one line'
    ' Grey*                scroll context window to found text'
    ''
end

/****************************************************************************\
     file loading routines
\****************************************************************************/

proc GetNameAndLine( var string name, var integer line )
    PushBlock()
    PushPosition()
    line = Val(GetText(1,6))
    if line
        lFind(HeaderLine,"bx")
    else
        lFind(HeaderLine,"cgx")
    endif
    MarkFoundText(1)
    name = GetMarkedText()
    PopPosition()
    PopBlock()
end

proc LoadFile()
    integer line
    string name[255] = ""

    GetNameAndLine(name,line)
    if Length(name) and  EditFile(QuotePath(name))
        GotoLine(line)
    endif
end

/****************************************************************************\
     context painting routine
\****************************************************************************/

proc PaintContext()
    integer id
    integer line
    integer win_line
    integer end_line = ctxt_start + ctxt_lines + 1
    string name[255] = ""

    GetNameAndLine(name,line)
    id = GetBufferId(name)
    if id
        GotoBufferId(id)
        PushPosition()
    else
        if not (Length(name) and name[1] <> "*"
                    and EditFile(QuotePath(name),_DONT_PROMPT_))
            return()
        endif
    endif

    win_line = line - ctxt_offset
    if win_line < 1
        win_line = 1
        ctxt_offset = line - 1
    elseif win_line > NumLines()
        win_line = NumLines()
        ctxt_offset = line - NumLines()
    endif

    GotoLine(win_line)

    name = SplitPath(name,_NAME_|_EXT_)
    if Length(name) > 20
        name = name[1..17] + "..."
    endif

#ifdef WIN32
    BufferVideo()
#endif

    Set(Attr,Query(MenuTextAttr))
    VGotoXY(1, ctxt_start)

#if EDITOR_VERSION <= 0x3000
    PutCharH("", Query(ScreenCols)-2)
    PutChar("")
#else
    PutOemCharH("", Query(ScreenCols)-2)
    PutOemChar("")
#endif

    VGotoXY(4,ctxt_start)
    PutStr(Format("[",line,"]"))
    VGotoXY(Query(ScreenCols)-Length(name)-6,ctxt_start)
    PutStr(Format("[",name,"]"))

    win_line = ctxt_start + 1
    repeat
        if line == CurrLine()
            Set(Attr,Query(MenuTextLtrAttr))
        else
            Set(Attr,Query(MenuTextAttr))
        endif
        VGotoXY(1,win_line)
        PutLine(GetText(1,CurrLineLen()),Query(ScreenCols)-2)
        win_line = win_line + 1
    until not Down() or win_line > end_line

    Set(Attr,Query(MenuTextAttr))
    while win_line <= end_line
        VGotoXY(1,win_line)
        PutCharH(" ",Query(ScreenCols)-2)
        win_line = win_line + 1
    endwhile

#ifdef WIN32
    UnBufferVideo()
#endif

    if id
        PopPosition()
    else
        AbandonFile()
    endif

    GotoBufferId(Query(ViewFindsId))
end

/****************************************************************************\
    list commands
\****************************************************************************/

proc LeftButton()
    integer delta = Query(MouseY) - CurrRow()

    if MouseHotSpot() == _MOUSE_MARKING_
        if delta > 0
            do delta times
                Down()
            enddo
        elseif delta < 0
            do -delta times
                Up()
            enddo
        else
            if GetClockTicks() - mouse_ticks < DOUBLE_CLICK_DELAY
                EndProcess(STOP_GOTO)
            endif
        endif
        mouse_ticks = GetClockTicks()
    elseif MouseHotSpot() == _MOUSE_CLOSE_
        EndProcess(STOP_QUIT)
    else
        ProcessHotSpot()
    endif
end

proc ScrollContextUp()
    idle_paint = FALSE
    ctxt_offset = ctxt_offset + 1
end

proc ScrollContextDown()
    idle_paint = FALSE
    ctxt_offset = ctxt_offset - 1
end

proc CenterContext()
    idle_paint = FALSE
    ctxt_offset = ctxt_lines / 2
end

#if EDITOR_VERSION <= 0x3000

proc MouseScrollUp()
    MouseStatus()
    if Query(MouseY) >= ctxt_start
        ScrollContextUp()
    else
        Up()
    endif
end

proc MouseScrollDown()
    MouseStatus()
    if Query(MouseY) >= ctxt_start
        ScrollContextDown()
    else
        Down()
    endif
end

#endif

integer proc PrevFileHeader()
    integer rc

    BegLine()
    rc = lFind(HeaderLine,"bx")
    ScrollToCenter()
    return(rc)
end

integer proc NextFileHeader()
    integer rc

    EndLine()
    rc = lFind(HeaderLine,"x")
    ScrollToCenter()
    return(rc)
end

proc LoadCurrFile()
    LoadFile()
    GotoBufferId(Query(ViewFindsId))
end

proc DeleteExprLines()
    while lFind(ResultLine,"gx")
        KillLine()
    endwhile
    BegFile()
end

proc DeleteThisEntry()
    PushBlock()
    if NextFileHeader()
        Up()
    else
        EndFile()
    endif
    MarkLine()
    PrevFileHeader()
    KillBlock()
    PopBlock()
end

proc KeepOnlyThisEntry()
    PushBlock()
    if NextFileHeader()
        MarkLine()
        EndFile()
        KillBlock()
    endif
    if PrevFileHeader() and Up()
        MarkLine()
        BegFile()
        KillBlock()
    endif
    PopBlock()
end

/****************************************************************************\
    list hook
\****************************************************************************/

#ifdef WIN32
constant _CtrlAlt_CursorUp              =   <CtrlAlt CursorUp>
constant _CtrlAlt_CursorDown            =   <CtrlAlt CursorDown>
#else
constant _CtrlAlt_CursorUp              =   <CtrlAlt GreyCursorUp>
constant _CtrlAlt_CursorDown            =   <CtrlAlt GreyCursorDown>
#endif

keydef ListKeys
    <LeftBtn>               LeftButton()

    <Escape>                EndProcess(STOP_QUIT)
    <Enter>                 EndProcess(STOP_GOTO)
    <Alt E>                 EndProcess(STOP_EDIT)
    <F5>                    EndProcess(STOP_GREP)
    <Alt C>                 EndProcess(STOP_CTXT)

    <F1>                    InitHelp(HLP_WIDTH) QuickHelp(ViewFindsHelp)
    <Alt L>                 LoadCurrFile()
    <Alt H>                 DeleteExprLines()
    <Alt R>                 DeleteThisEntry()
    <Alt T>                 KeepOnlyThisEntry()

    <Ctrl CursorUp>         PrevFileHeader()
    <Ctrl CursorDown>       NextFileHeader()

    <Ctrl L>                RepeatFind()

    <Grey->                 ScrollContextUp()
    <Grey+>                 ScrollContextDown()
    <Grey*>                 CenterContext()

#if EDITOR_VERSION <= 0x3000
    <_CtrlAlt_CursorUp>     MouseScrollUp()
    <_CtrlAlt_CursorDown>   MouseScrollDown()
#endif

end

proc ListStartup()
    ListFooter(" {F1}-Help ")
    DisplayMode(_DISPLAY_FINDS_)
    Enable(ListKeys)
    UnHook(ListStartup)
    BreakHookChain()
end

proc ListCleanup()
    DisplayMode(_DISPLAY_TEXT_)
    UnHook(ListCleanup)
end

/****************************************************************************\
    idle handling (paint context window)
\****************************************************************************/

proc IdleProc()
    if GetClockTicks() >= idle_ticks
        PaintContext()
    endif
end

proc ListCommand()
    if idle_paint
        ctxt_offset = ctxt_lines / 2
        idle_ticks = GetClockTicks() + IDLE_UPDATE_DELAY
    else
        PaintContext()
        idle_paint = TRUE
    endif
end

/****************************************************************************\
    termination routines
\****************************************************************************/

proc GotoCurrFile()
    LoadFile()
    ScrollToCenter()
end

proc EditViewBuffer()
    integer count = 0

    BufferType(_NORMAL_)
    repeat
        count = count + 1
#if EDITOR_VERSION >= 0x3000
    until ChangeCurrFilename(Format("<unnamed-",count,">"),_DONT_PROMPT_)
#else
    until ChangeCurrFilename(Format("[unnamed-",count,"]"),_DONT_PROMPT_)
#endif
    Set(ViewFindsId,0)
end

proc LimitContextLines()
    if ctxt_lines < 4
        ctxt_lines = 0
    elseif ctxt_lines > Query(ScreenRows) - 10
        ctxt_lines = Query(ScreenRows) - 10
    endif
end

proc ChangeContextWindow()
    string reply[8] = Str(ctxt_lines+1)

    Set(X1,1)
    Set(Y1,1)
    AskNumeric("Height of Context Window",reply,hist_lines)
    ctxt_lines = Val(reply) - 1
    LimitContextLines()
end

/****************************************************************************\
    main program
\****************************************************************************/

proc Main()
    integer rc, ctxt
    integer lst_height
    integer bid = GetBufferId()

    // start up

    hist_lines = GetFreeHistory("DlgGrep:ContextSize")
    ctxt_lines = Val(GetHistoryStr(hist_lines,1)) - 1
    LimitContextLines()

restart:

    // check for view buffer

    if not GotoBufferId(Query(ViewFindsId))
        Warn("View buffer does not exist")
        goto common_exit
    endif

    // open context window

    if ctxt_lines == 0
        ctxt = FALSE
        lst_height = Query(ScreenRows) - 2
    else
        lst_height = Query(ScreenRows) - ctxt_lines - 5
        ctxt_start = lst_height + 2
        ctxt_offset = ctxt_lines / 2

        ctxt = PopWinOpen(1,ctxt_start+1,Query(ScreenCols),
                Query(ScreenRows),1,"Context",Query(MenuTextAttr))

        if ctxt
            Set(Attr,Query(MenuTextAttr))
            ClrScr()
            idle_paint = TRUE
            idle_ticks = GetClockTicks()
            Hook(_AFTER_NONEDIT_COMMAND_,ListCommand)
            Hook(_NONEDIT_IDLE_,IdleProc)
        endif
    endif

    // run list command

    Hook(_LIST_STARTUP_,ListStartup)
    Hook(_LIST_CLEANUP_,ListCleanup)

    rc = lList("ViewFinds",Query(ScreenCols),lst_height,
          _FIXED_HEIGHT_|_ENABLE_SEARCH_|_ENABLE_HSCROLL_)

    // close context window

    if ctxt
        PopWinClose()
        UnHook(IdleProc)
        UnHook(ListCommand)
    endif

    // handle result of list command

    case rc
        when STOP_GOTO
            GotoCurrFile()
        when STOP_EDIT
            EditViewBuffer()
        when STOP_GREP
            GotoBufferId(bid)
            ExecMacro("grep refresh")
            goto restart
        when STOP_CTXT
            ChangeContextWindow()
            goto restart
        otherwise
            GotoBufferId(bid)
    endcase

    // clean up

common_exit:

    UpdateDisplay(_ALL_WINDOWS_REFRESH_)
    PurgeMacro(CurrMacroFilename())
end
